import { isRef, ref, computed, watch, onUnmounted, reactive, inject, provide } from '@vue/composition-api';

function unwrap(o) {
    return isRef(o) ? o.value : o;
}
function wrap(o) {
    return isRef(o) ? o : ref(o);
}
const isArray = Array.isArray;
const isFunction = (val) => typeof val === "function";
const isString = (val) => typeof val === "string";
const isSymbol = (val) => typeof val === "symbol";
const isBoolean = (val) => typeof val === "boolean";
const isDate = (val) => isObject(val) && isFunction(val.getTime);
const isNumber = (val) => typeof val === "number";
const isObject = (val) => val !== null && typeof val === "object";
const isElement = (val) => isObject(val) && !!val.tagName;
function isPromise(val) {
    return isObject(val) && isFunction(val.then) && isFunction(val.catch);
}
const NO_OP = () => { };
const FALSE_OP = () => false;
const PASSIVE_EV = { passive: true };
function promisedTimeout(timeout) {
    return new Promise(res => {
        setTimeout(res, timeout);
    });
}
function minMax(val, min, max) {
    if (val < min)
        return min;
    if (val > max)
        return max;
    return val;
}
const isClient = typeof window != "undefined";
function deepClone(result, ...sources) {
    for (let i = 0; i < sources.length; i++) {
        const source = sources[i];
        if (source === undefined || !isObject(source))
            continue;
        const keys = Object.keys(source);
        for (let j = 0; j < keys.length; j++) {
            const k = keys[j];
            const v = unwrap(source[k]);
            const sourceType = typeof v;
            const type = typeof result[k];
            if (result[k] === undefined || sourceType === type) {
                result[k] = isObject(v)
                    ? deepClone(result[k] || {}, v)
                    : source[k]; // source[k] is assigned because if is ref we want to override to this ref
            }
        }
    }
    return result;
}
// compact version: https://stackoverflow.com/a/33146982/1209882
/**
 * returns a random string
 * @param len length of the string max: 36
 */
// export function randomString(len: number) {
//   return (+new Date).toString(36).slice(-len);
// }

function usePagination(options) {
    const _currentPage = wrap(options.currentPage);
    const _pageSize = wrap(options.pageSize);
    const _offset = ref(0);
    const total = wrap(options.total);
    const offset = computed({
        get() {
            return _offset.value;
        },
        set(v) {
            if (!isNumber(v)) {
                /* istanbul ignore else */
                if ((process.env.NODE_ENV !== 'production')) {
                    console.warn(`[offset] expected number but got: '${typeof v}' value: '${v}'`);
                }
                return;
            }
            _offset.value = Math.min(v, total.value);
        }
    });
    const currentPage = computed({
        get() {
            return _currentPage.value;
        },
        set(v) {
            if (!isNumber(v)) {
                /* istanbul ignore else  */
                if ((process.env.NODE_ENV !== 'production')) {
                    console.warn(`[currentPage] expected number but got: '${typeof v}' value: '${v}'`);
                }
                return;
            }
            _currentPage.value = minMax(v, 1, lastPage.value);
            // set the offset
            offset.value = (_currentPage.value - 1) * pageSize.value;
        }
    });
    const pageSize = computed({
        get() {
            return _pageSize.value;
        },
        set(v) {
            if (!isNumber(v)) {
                /* istanbul ignore else */
                if ((process.env.NODE_ENV !== 'production')) {
                    console.warn(`[pageSize] expected number but got: '${typeof v}' value: '${v}'`);
                }
                return;
            }
            _pageSize.value = v;
        }
    });
    const lastPage = computed(() => Math.ceil(total.value / pageSize.value));
    // make sure the current page is the correct value
    currentPage.value = _currentPage.value;
    const prev = () => --currentPage.value;
    const next = () => ++currentPage.value;
    const first = () => (currentPage.value = 1);
    const last = () => (currentPage.value = lastPage.value);
    watch([total, pageSize], () => {
        if (currentPage.value > lastPage.value) {
            currentPage.value = lastPage.value;
        }
    }, { lazy: true } // no need to run on first render
    );
    return {
        // Mutable state
        pageSize,
        total,
        currentPage,
        offset,
        // Computed
        lastPage,
        // Functions
        next,
        prev,
        first,
        last
    };
}

function useArrayPagination(array, options) {
    const arrayRef = wrap(array);
    const pagination = usePagination({
        ...{
            currentPage: 1,
            pageSize: 10
        },
        ...options,
        total: computed(() => arrayRef.value.length)
    });
    const result = computed(() => {
        const array = arrayRef.value;
        if (!Array.isArray(array))
            return [];
        return array.slice(pagination.offset.value, pagination.offset.value + pagination.pageSize.value);
    });
    return {
        ...pagination,
        result
    };
}

function usePromise(fn, throwException = false) {
    if (!fn) {
        throw new Error(`[usePromise] argument can't be '${fn}'`);
    }
    if (typeof fn !== "function") {
        throw new Error(`[usePromise] expects function, but received ${typeof fn}`);
    }
    const loading = ref(false);
    const error = ref(null);
    const result = ref(null);
    const promise = ref();
    const exec = async (...args) => {
        loading.value = true;
        error.value = null;
        result.value = null;
        let throwExp = args &&
            fn.length !== args.length &&
            args.length > 0 &&
            isBoolean(args[args.length - 1])
            ? args[args.length - 1]
            : throwException;
        const currentPromise = (promise.value = fn(...args));
        try {
            const r = await currentPromise;
            if (promise.value === currentPromise) {
                result.value = r;
            }
            return r;
        }
        catch (er) {
            if (promise.value === currentPromise) {
                error.value = er;
                result.value = null;
            }
            return throwExp ? currentPromise : undefined;
        }
        finally {
            if (promise.value === currentPromise) {
                loading.value = false;
            }
        }
    };
    return {
        exec,
        result,
        promise,
        loading,
        error
    };
}

function useCancellablePromise(fn, throwException = false) {
    const cancelled = ref(false);
    let _cancel = undefined;
    const cancel = (result) => _cancel(result); // TODO add warning if cancel is undefined
    const promise = (p) => new Promise((res, rej) => {
        _cancel = result => {
            cancelled.value = true;
            rej(result);
        };
        p.then(res).catch(rej);
    });
    const use = usePromise((...args) => promise(fn(...args)), throwException);
    return {
        ...use,
        cancel,
        cancelled
    };
}

const MAX_RETRIES = 9000;
/* istanbul ignore next */
const ExecutionId = /*#__PURE__*/ Symbol(((process.env.NODE_ENV !== 'production') && "RetryId") || ``);
/* istanbul ignore next */
const CancellationToken = /*#__PURE__*/ Symbol(((process.env.NODE_ENV !== 'production') && "CancellationToken") || ``);
const defaultStrategy = async (options, context, factory, args) => {
    const retryId = context[ExecutionId].value;
    let i = -1;
    const maxRetries = options.maxRetries || MAX_RETRIES + 1;
    const delay = options.retryDelay || noDelay;
    context.retryErrors.value = [];
    context.isRetrying.value = false;
    context.nextRetry.value = undefined;
    let nextRetry = undefined;
    do {
        let success = false;
        let result = null;
        try {
            ++i;
            if (args) {
                result = factory(...args);
            }
            else {
                result = factory();
            }
            if (isPromise(result)) {
                result = await result;
            }
            // is cancelled?
            if (context[CancellationToken].value) {
                return null;
            }
            success = true;
        }
        catch (error) {
            result = null;
            context.retryErrors.value.push(error);
        }
        // is our retry current one?
        if (retryId !== context[ExecutionId].value) {
            return result;
        }
        if (success) {
            context.isRetrying.value = false;
            context.nextRetry.value = undefined;
            return result;
        }
        if (i >= maxRetries) {
            context.isRetrying.value = false;
            context.nextRetry.value = undefined;
            return Promise.reject(new Error(`[useRetry] max retries reached ${maxRetries}`));
        }
        context.isRetrying.value = true;
        const now = Date.now();
        const pDelayBy = delay(i); // wrapped
        const delayBy = isPromise(pDelayBy) ? await pDelayBy : pDelayBy; // unwrap promise
        if (!isPromise(pDelayBy) || !!delayBy) {
            if (isNumber(delayBy)) {
                nextRetry = delayBy;
            }
            else if (isDate(delayBy)) {
                nextRetry = delayBy.getTime();
            }
            else {
                throw new Error(`[useRetry] invalid value received from options.retryDelay '${typeof delayBy}'`);
            }
            // if the retry is in the past, means we need to await that amount
            if (nextRetry < now) {
                context.nextRetry.value = now + nextRetry;
            }
            else {
                context.nextRetry.value = nextRetry;
                nextRetry = nextRetry - now;
            }
            if (nextRetry > 0) {
                await promisedTimeout(nextRetry);
            }
        }
        // is cancelled?
        if (context[CancellationToken].value) {
            return null;
        }
        // is our retry current one?
        if (retryId !== context[ExecutionId].value) {
            return result;
        }
    } while (i < MAX_RETRIES);
    return null;
};
function useRetry(options, factory) {
    const opt = !options || isFunction(options) ? {} : options;
    const fn = isFunction(options) ? options : factory;
    if (!isFunction(options) && !isObject(options)) {
        throw new Error("[useRetry] options needs to be 'object'");
    }
    if (!!fn && !isFunction(fn)) {
        throw new Error("[useRetry] factory needs to be 'function'");
    }
    const isRetrying = ref(false);
    const nextRetry = ref();
    const retryErrors = ref([]);
    const cancellationToken = { value: false };
    const retryId = { value: 0 };
    const retryCount = computed(() => retryErrors.value.length);
    const context = {
        isRetrying,
        retryCount,
        nextRetry,
        retryErrors,
        [ExecutionId]: retryId,
        [CancellationToken]: cancellationToken
    };
    const exec = fn
        ? (...args) => {
            ++context[ExecutionId].value;
            return defaultStrategy(opt, context, fn, args);
        }
        : (f) => {
            ++context[ExecutionId].value;
            return defaultStrategy(opt, context, f, undefined);
        };
    const cancel = () => {
        context.isRetrying.value = false;
        context.retryErrors.value.push(new Error("[useRetry] cancelled"));
        context.nextRetry.value = undefined;
        cancellationToken.value = true;
    };
    return {
        ...context,
        cancel,
        exec
    };
}
const exponentialDelay = retryNumber => {
    const delay = Math.pow(2, retryNumber) * 100;
    const randomSum = delay * 0.2 * Math.random(); // 0-20% of the delay
    return delay + randomSum;
};
const noDelay = () => 0;

/**
 * Create `debounced` function, options object can be changed after creation to update behaviour
 * @param handler - function to be debounced
 * @param wait - debounce ms
 * @param options - debounce options
 */
function useDebounce(handler, wait, options) {
    return debounce(handler, wait, options);
}
function debounce(func, waitMilliseconds = 50, options = {
    isImmediate: false
}) {
    let timeoutId;
    return function (...args) {
        const context = this;
        const doLater = function () {
            timeoutId = undefined;
            if (!options.isImmediate) {
                func.apply(context, args);
            }
        };
        const shouldCallNow = options.isImmediate && timeoutId === undefined;
        if (timeoutId !== undefined) {
            clearTimeout(timeoutId);
        }
        timeoutId = setTimeout(doLater, waitMilliseconds);
        if (shouldCallNow) {
            func.apply(context, args);
        }
    };
}

/**
 * Reactive custom timer with specified refresh rate
 * @param options - Configuration
 */
function useNow(options) {
    const SYNC_MS = 1000;
    const ms = (options && options.refreshMs) || SYNC_MS;
    const sync = options && isBoolean(options.sync) ? options.sync : true;
    const fn = (options && isFunction(options.timeFn) && options.timeFn) || Date.now;
    /* istanbul ignore else */
    if ((process.env.NODE_ENV !== 'production')) {
        if (options && options.timeFn && isFunction(options.timeFn) === false) {
            console.warn("[useNow] timeFn param must be Function");
        }
    }
    let handler = undefined;
    let timeoutHandler = undefined;
    const now = ref(fn());
    const remove = () => {
        clearInterval(handler);
        clearTimeout(timeoutHandler);
    };
    /* istanbul ignore next */
    const start = isClient
        ? () => (handler = setInterval(() => (now.value = fn()), ms))
        : NO_OP;
    if (sync) {
        const offset = SYNC_MS - (now.value - Math.floor(now.value / SYNC_MS) * SYNC_MS);
        timeoutHandler = setTimeout(start, offset);
    }
    else {
        start();
    }
    onUnmounted(remove);
    return {
        now,
        remove
    };
}

/**
 * Reactive `Date.now()` with specified refresh rate
 * @param options - Configuration
 */
function useDateNow(options) {
    const refreshMs = (options && options.refreshMs) || 1000;
    const sync = options && isBoolean(options.sync) ? options.sync : true;
    return useNow({
        refreshMs,
        sync,
        timeFn: Date.now
    });
}

/**
 * Reactive `performance.now()` with specified refresh rate
 * @param options - Configuration
 */
function usePerformanceNow(options) {
    const refreshMs = (options && options.refreshMs) || 1000;
    const sync = options && isBoolean(options.sync) ? options.sync : true;
    return useNow({
        refreshMs,
        sync,
        timeFn: () => performance.now()
    });
}

function useFormat(format, args) {
    return computed(() => {
        const f = unwrap(format);
        if (!args) {
            return f;
        }
        const r = isRef(args) ? reactive(args.value) : reactive(args);
        const regEx = /({?{[\w\s]*}?})/g;
        return f.replace(regEx, s => {
            const k = s
                .replace("{", "")
                .replace("}", "")
                .trim();
            // ignore if {{ }}
            if (s[0] === s[1] && s[0] === "{") {
                return `${k}`;
            }
            const v = isArray(r) ? r[+k] : r[k];
            return v === undefined ? s : `${unwrap(v)}`;
        });
    });
}

function usePath(source, path, separator = ".", notFoundReturn = NO_OP) {
    return computed(() => {
        const s = unwrap(source);
        const p = unwrap(path);
        if (s === undefined)
            return notFoundReturn(p, s, p, s);
        if (!p) {
            return s;
        }
        const fragments = p.split(separator);
        let c = s;
        for (let i = 0; i < fragments.length; i++) {
            let fragmentPath = fragments[i];
            let index = -1;
            if (fragmentPath[fragmentPath.length - 1] === "]") {
                const m = fragmentPath.match(/\[(\d+)\]$/);
                if (m && m[1]) {
                    index = +m[1];
                    fragmentPath = fragmentPath.slice(0, -m[0].length);
                }
            }
            if (isObject(c)) {
                c = c[fragmentPath];
                // array like: when using ref with and array, it becomes arraylike object
                if (index >= 0) {
                    c = c[index];
                }
            }
            else {
                if ((process.env.NODE_ENV !== 'production')) {
                    console.warn(`Path "${fragments.slice(0, i).join(separator)}" doesn't exist on:`, source);
                }
                return notFoundReturn(fragments.slice(0, i).join(separator), c, p, s);
            }
            if (!c) {
                return notFoundReturn(fragments.slice(0, i).join(separator), c, p, s);
            }
        }
        return c;
    });
}

/* /Output */
function isValidation(v) {
    return typeof v.$value !== "undefined";
}
function isValidatorObject(v) {
    return isObject(v);
}
const buildValidationFunction = (r, f, handlers) => {
    const $promise = ref(null);
    const $pending = ref(false);
    const $error = ref();
    const $invalid = ref(false);
    let context = undefined;
    const onChange = (r) => {
        const p = async () => {
            try {
                $pending.value = true;
                const result = f(r, context);
                if (isPromise(result)) {
                    $invalid.value = !(await result);
                }
                else {
                    $invalid.value = !result;
                }
            }
            catch (e) {
                $invalid.value = true;
                throw e;
            }
            finally {
                $pending.value = false;
            }
        };
        $promise.value = p().catch(x => {
            $error.value = x;
            $invalid.value = true;
            return x;
        });
    };
    handlers.push((ctx) => {
        context = ctx;
        watch(() => {
            try {
                // keep track on the external dependencies
                f(r.value, context);
            }
            catch (e) {
                // ignore error
            }
            return r.value;
        }, onChange, { deep: true });
    });
    return {
        $promise,
        $pending,
        $invalid,
        $error
    };
};
const buildValidationValue = (r, v, handlers) => {
    const { $message, $validator, ...$rest } = isValidatorObject(v)
        ? v
        : { $validator: v, $message: "" };
    const { $pending, $promise, $invalid, $error } = buildValidationFunction(r, $validator, handlers);
    return {
        $pending,
        $error,
        $promise,
        $invalid,
        $message,
        ...$rest
    };
};
const buildValidation = (o, handlers) => {
    const r = {};
    const $value = isValidation(o) ? o.$value : undefined;
    for (const k of Object.keys(o)) {
        if (k[0] === "$") {
            if (k === "$value") {
                r[k] = $value;
                const $dirty = ref(false);
                const dirtyWatch = watch($value, () => {
                    $dirty.value = true;
                    dirtyWatch();
                }, { lazy: true, deep: true });
                r["$dirty"] = $dirty;
                continue;
            }
            else {
                r[k] = o[k];
                continue;
            }
        }
        if ($value) {
            const validation = buildValidationValue($value, o[k], handlers);
            r[k] = {
                ...validation,
                $value
            };
        }
        else {
            const validation = buildValidation(o[k], handlers);
            let $anyDirty = undefined;
            let $errors;
            let $anyInvalid;
            if (isValidation(validation)) {
                const validations = Object.keys(validation)
                    .filter(x => x[0] !== "$")
                    .map(x => validation[x]);
                $errors = computed(() => {
                    return validations
                        .map(x => x.$error)
                        .filter(x => Boolean(x) && Array.isArray(x) ? x.some(Boolean) : x);
                });
                // $anyDirty = computed(() => validations.some(x => !!x));
                $anyInvalid = computed(() => validations.some(x => !!x.$invalid));
            }
            else {
                const validations = Object.keys(validation).map(x => validation[x]);
                $errors = computed(() => {
                    return validations
                        .map(x => x.$errors)
                        .filter(Boolean)
                        .filter(x => x.some(Boolean));
                });
                $anyDirty = computed(() => validations.some(x => (x.$anyDirty && x.$anyDirty) ||
                    (isBoolean(x.$dirty) && x.$dirty)));
                $anyInvalid = computed(() => validations.some(x => !!x.$anyInvalid));
            }
            r[k] = {
                ...validation,
                $errors,
                $anyInvalid
            };
            if ($anyDirty) {
                r[k].$anyDirty = $anyDirty;
            }
        }
    }
    return r;
};
function useValidation(input) {
    const handlers = [];
    const validation = buildValidation({ input }, handlers);
    // convert to reactive, this will make it annoying to deconstruct, but
    // allows to use it directly on the render template without `.value`
    // https://github.com/vuejs/vue-next/pull/738
    const validationInput = reactive(validation.input);
    // set the context, this will allow to use this object as the second
    // argument when calling validators
    handlers.forEach(x => x(validationInput));
    return validationInput;
}

const I18n_ACCESS_SYMBOL = Symbol(((process.env.NODE_ENV !== 'production') && "I18n") || ``);
function useI18n(definition) {
    if (definition) {
        return buildI18n(definition);
    }
    else
        return inject(I18n_ACCESS_SYMBOL);
}
function buildI18n(definition) {
    const locales = ref(Object.keys(definition.messages));
    const localeMessages = ref(definition.messages);
    const locale = ref(definition.locale);
    const i18n = ref({});
    let fallback = ref();
    const cache = {};
    const loadLocale = (locale, localeMessages) => {
        if (cache[locale]) {
            return cache[locale];
        }
        const l = localeMessages.value[locale];
        if (!l) {
            return ref({});
        }
        if (isFunction(l)) {
            return Promise.resolve(l()).then(x => (cache[locale] = wrap(x)));
        }
        return (cache[locale] = computed(() => localeMessages.value[locale]));
    };
    const shouldFallback = definition.fallback
        ? isBoolean(definition.notFoundFallback)
            ? definition.notFoundFallback
            : true
        : false;
    let fallbackIsPromise = false;
    if (shouldFallback) {
        const fallbackI18n = loadLocale(definition.fallback, localeMessages);
        if (isPromise(fallbackI18n)) {
            fallbackI18n.then(x => {
                fallback = x;
            });
            fallbackIsPromise = true;
        }
        else {
            fallback = fallbackI18n;
        }
    }
    else {
        fallback.value = {};
    }
    watch([locale, fallback], async ([l, fb]) => {
        if (l === definition.fallback && shouldFallback) {
            i18n.value = fb;
        }
        else {
            const localeMessage = await loadLocale(l, localeMessages);
            i18n.value = deepClone({}, fb, localeMessage.value);
        }
    }, {
        lazy: fallbackIsPromise
    });
    const $t = (path, args) => {
        if (definition.resolve) {
            return wrap(definition.resolve(i18n.value, path, args));
        }
        return useFormat(usePath(i18n, path, ".", (_, _1, p) => p), args);
    };
    const addLocale = (l, m) => {
        if (locales.value.indexOf(l) >= 0) {
            if ((process.env.NODE_ENV !== 'production')) {
                console.warn("Locale already exists, overriding it");
            }
        }
        else {
            locales.value.push(l);
        }
        delete cache[l];
        localeMessages.value[l] = m;
    };
    const removeLocale = (l) => {
        const index = locales.value.indexOf(l);
        if (index >= 0) {
            locales.value.splice(index, 1);
        }
        else {
            if ((process.env.NODE_ENV !== 'production')) {
                console.warn("Locale doesn't exist");
            }
        }
        delete localeMessages.value[l];
        delete cache[l];
    };
    return {
        locale,
        locales,
        i18n,
        $t,
        addLocale,
        removeLocale
    };
}
function setI18n(definition) {
    const r = buildI18n(definition);
    provide(I18n_ACCESS_SYMBOL, r);
    return r;
}

export { FALSE_OP, NO_OP, PASSIVE_EV, buildI18n, debounce, deepClone, exponentialDelay, isArray, isBoolean, isClient, isDate, isElement, isFunction, isNumber, isObject, isPromise, isString, isSymbol, minMax, noDelay, promisedTimeout, setI18n, unwrap, useArrayPagination, useCancellablePromise, useDateNow, useDebounce, useFormat, useI18n, useNow, usePagination, usePath, usePerformanceNow, usePromise, useRetry, useValidation, wrap };
